@@ -1,11 +1,11 @@
 import argparse
-from collections import defaultdict
 import logging
-from typing import Dict, List, Set, Tuple
+from typing import Dict, List
 
 from mir.commands import base
 from mir.protos import mir_command_pb2 as mirpb
-from mir.tools import checker, class_ids, revs_parser, mir_repo_utils, mir_storage, mir_storage_ops
+from mir.tools import checker, class_ids, revs_parser, mir_repo_utils, mir_storage_ops
+from mir.tools.annotations import map_and_filter_annotations
 from mir.tools.code import MirCode
 from mir.tools.command_run_in_out import command_run_in_out
 from mir.tools.errors import MirRuntimeError
@@ -17,8 +17,10 @@ class CmdCopy(base.BaseCommand):
     def run(self) -> int:
         logging.debug("command copy: %s", self.args)
         return CmdCopy.run_with_args(mir_root=self.args.mir_root,
+                                     label_storage_file=self.args.label_storage_file,
                                      data_mir_root=self.args.data_mir_root,
                                      data_src_revs=self.args.data_src_revs,
+                                     data_label_storage_file=self.args.data_label_storage_file,
                                      dst_rev=self.args.dst_rev,
                                      ignore_unknown_types=self.args.ignore_unknown_types,
                                      drop_annotations=self.args.drop_annotations,
@@ -28,8 +30,10 @@ class CmdCopy(base.BaseCommand):
     @staticmethod
     @command_run_in_out
     def run_with_args(mir_root: str,
+                      label_storage_file: str,
                       data_mir_root: str,
                       data_src_revs: str,
+                      data_label_storage_file: str,
                       dst_rev: str,
                       ignore_unknown_types: bool,
                       drop_annotations: bool,
@@ -52,63 +56,58 @@ class CmdCopy(base.BaseCommand):
                                                task_name=dst_typ_rev_tid.tid)
 
         check_code = checker.check(mir_root,
-                                   [checker.Prerequisites.IS_INSIDE_MIR_REPO, checker.Prerequisites.HAVE_LABELS])
+                                   [checker.Prerequisites.IS_INSIDE_MIR_REPO])
         if check_code != MirCode.RC_OK:
             return check_code
 
         check_code = checker.check(
-            data_mir_root, prerequisites=[checker.Prerequisites.IS_INSIDE_MIR_REPO, checker.Prerequisites.HAVE_LABELS])
+            data_mir_root, prerequisites=[checker.Prerequisites.IS_INSIDE_MIR_REPO])
         if check_code != MirCode.RC_OK:
             return check_code
 
+        PhaseLoggerCenter.update_phase(phase="copy.init")
+
         # read from src mir root
-        [mir_metadatas, mir_annotations, mir_keywords, mir_tasks,
-         _] = mir_storage_ops.MirStorageOps.load_multiple_storages(mir_root=data_mir_root,
-                                                                   mir_branch=data_src_typ_rev_tid.rev,
-                                                                   mir_task_id=data_src_typ_rev_tid.tid,
-                                                                   ms_list=mir_storage.get_all_mir_storage(),
-                                                                   as_dict=False)
+        mir_metadatas: mirpb.MirMetadatas
+        mir_annotations: mirpb.MirAnnotations
+        mir_tasks: mirpb.MirTasks
+        mir_metadatas, mir_annotations, mir_tasks = mir_storage_ops.MirStorageOps.load_multiple_storages(
+            mir_root=data_mir_root,
+            mir_branch=data_src_typ_rev_tid.rev,
+            mir_task_id=data_src_typ_rev_tid.tid,
+            ms_list=[mirpb.MIR_METADATAS, mirpb.MIR_ANNOTATIONS, mirpb.MIR_TASKS],
+            as_dict=False)
 
         PhaseLoggerCenter.update_phase(phase='copy.read')
 
-        orig_head_task_id = mir_annotations.head_task_id
-        if not orig_head_task_id:
-            logging.error('bad annotations.mir: empty head task id')
-            return MirCode.RC_CMD_INVALID_MIR_REPO
-        if ((len(mir_annotations.task_annotations) > 0 and orig_head_task_id not in mir_annotations.task_annotations)):
-            logging.error(f"bad annotations.mir: can not find head task id: {orig_head_task_id}")
-            return MirCode.RC_CMD_INVALID_MIR_REPO
+        unknown_class_names: List[str] = []
+        if drop_annotations:
+            mir_annotations.prediction.Clear()
+            mir_annotations.prediction.type = mirpb.ObjectType.OT_NO_ANNOTATIONS
+            mir_annotations.ground_truth.Clear()
+            mir_annotations.ground_truth.type = mirpb.ObjectType.OT_NO_ANNOTATIONS
+        else:
+            unknown_class_names = map_and_filter_annotations(mir_annotations=mir_annotations,
+                                                             data_label_storage_file=data_label_storage_file,
+                                                             label_storage_file=label_storage_file)
 
-        single_task_annotations, unknown_types = CmdCopy._change_single_task_annotations(
-            data_mir_root=data_mir_root,
-            dst_mir_root=mir_root,
-            single_task_annotations=mir_annotations.task_annotations[orig_head_task_id],
-            ignore_unknown_types=ignore_unknown_types,
-            drop_annotations=drop_annotations)
-
-        mir_annotations.task_annotations[dst_typ_rev_tid.tid].CopyFrom(single_task_annotations)
-        del mir_annotations.task_annotations[orig_head_task_id]
-        mir_annotations.head_task_id = dst_typ_rev_tid.tid
-
-        # tasks.mir: get necessary head task infos, remove others and change head task id
-        orig_head_task_id = mir_tasks.head_task_id
-        if not orig_head_task_id:
-            logging.error('bad tasks.mir: empty head task id')
-            return MirCode.RC_CMD_INVALID_MIR_REPO
-        if orig_head_task_id not in mir_tasks.tasks:
-            logging.error(f"bad tasks.mir: can not find head task id: {orig_head_task_id}")
-            return MirCode.RC_CMD_INVALID_MIR_REPO
+        if unknown_class_names:
+            if ignore_unknown_types:
+                logging.warning(f"unknown types: {unknown_class_names}")
+            else:
+                raise MirRuntimeError(
+                    error_code=MirCode.RC_CMD_INVALID_MIR_REPO,
+                    error_message=f"copy annotations error, unknown class names: {unknown_class_names}")
 
         PhaseLoggerCenter.update_phase(phase='copy.change')
 
         # save and commit
-        orig_task = mir_tasks.tasks[orig_head_task_id]
+        orig_task: mirpb.Task = mir_tasks.tasks[mir_tasks.head_task_id]
         task = mir_storage_ops.create_task(task_type=mirpb.TaskType.TaskTypeCopyData,
                                            task_id=dst_typ_rev_tid.tid,
                                            message=f"copy from {data_mir_root}, src: {data_src_revs}, dst: {dst_rev}",
-                                           unknown_types=unknown_types,
-                                           model_hash=orig_task.model.model_hash,
-                                           model_mAP=orig_task.model.mean_average_precision,
+                                           new_types={name: 0 for name in unknown_class_names},
+                                           model_meta=orig_task.model,
                                            serialized_task_parameters=orig_task.serialized_task_parameters,
                                            serialized_executor_config=orig_task.serialized_executor_config,
                                            executor=orig_task.executor,
@@ -128,67 +127,41 @@ class CmdCopy(base.BaseCommand):
     @staticmethod
     def _change_type_ids(
         single_task_annotations: mirpb.SingleTaskAnnotations,
-        data_mir_root: str,
-        dst_mir_root: str,
-    ) -> Tuple[int, Dict[str, int]]:
-        src_to_dst_ids: Dict[int, int] = {}
-        unknown_types_and_count: Dict[str, int] = defaultdict(int)
-        dst_class_id_mgr = class_ids.ClassIdManager(mir_root=dst_mir_root)
-        src_class_id_mgr = class_ids.ClassIdManager(mir_root=data_mir_root)
-
-        for asset_id, single_image_annotations in single_task_annotations.image_annotations.items():
-            dst_keyids_set: Set[int] = set()
-            dst_image_annotations: List[mirpb.Annotation] = []
-            for annotation in single_image_annotations.annotations:
-                src_type_id = annotation.class_id
-                if not src_class_id_mgr.has_id(src_type_id):
-                    # if we can not find src type id in data_mir_root's labels.csv, this repo in invalid and cannot copy
-                    logging.error(f"broken data_mir_root, unknown src id: {annotation.class_id}")
-                    return MirCode.RC_CMD_INVALID_MIR_REPO, unknown_types_and_count
-
-                if src_type_id in src_to_dst_ids:
-                    # get mapping from cache
-                    annotation.class_id = src_to_dst_ids[src_type_id]
+        src_to_dst_ids: Dict[int, int],
+    ) -> None:
+        for single_image_annotations in single_task_annotations.image_annotations.values():
+            dst_image_annotations: List[mirpb.ObjectAnnotation] = []
+            for annotation in single_image_annotations.boxes:
+                dst_class_id = src_to_dst_ids[annotation.class_id]
+                if dst_class_id >= 0:
+                    annotation.class_id = dst_class_id
                     dst_image_annotations.append(annotation)
-                else:
-                    # if no cache, src_id -> src_type_name -> dst_id
-                    src_type_name = src_class_id_mgr.main_name_for_id(src_type_id) or ''
-                    if dst_class_id_mgr.has_name(src_type_name):
-                        annotation.class_id = dst_class_id_mgr.id_and_main_name_for_name(src_type_name)[0]
-                        dst_image_annotations.append(annotation)
+            del single_image_annotations.boxes[:]
+            single_image_annotations.boxes.extend(dst_image_annotations)
 
-                        src_to_dst_ids[src_type_id] = annotation.class_id  # save cache
-                        dst_keyids_set.add(annotation.class_id)
-                    else:
-                        unknown_types_and_count[src_type_name] += 1
-
-            del single_image_annotations.annotations[:]
-            single_image_annotations.annotations.extend(dst_image_annotations)
-        return MirCode.RC_OK, unknown_types_and_count
+        dst_eval_class_ids: List[int] = []
+        for src_class_id in single_task_annotations.eval_class_ids:
+            dst_class_id = src_to_dst_ids[src_class_id]
+            if dst_class_id >= 0:
+                dst_eval_class_ids.append(dst_class_id)
+        single_task_annotations.eval_class_ids[:] = dst_eval_class_ids
 
     @staticmethod
-    def _change_single_task_annotations(data_mir_root: str, dst_mir_root: str,
-                                        single_task_annotations: mirpb.SingleTaskAnnotations,
-                                        ignore_unknown_types: bool,
-                                        drop_annotations: bool) -> Tuple[mirpb.SingleTaskAnnotations, Dict[str, int]]:
-        if drop_annotations:
-            return mirpb.SingleTaskAnnotations(), {}
+    def _gen_unknown_names_and_count(src_class_id_mgr: class_ids.UserLabels, mir_context: mirpb.MirContext,
+                                     src_to_dst_ids: Dict[int, int]) -> Dict[str, int]:
+        all_src_class_ids = set(mir_context.pred_stats.class_ids_cnt.keys()) | set(
+            mir_context.gt_stats.class_ids_cnt.keys())
+        unknown_src_class_ids = {src_id for src_id in all_src_class_ids if src_to_dst_ids[src_id] == -1}
+        if not unknown_src_class_ids:
+            return {}
 
-        # if don't want to drop annotations
-        # annotations.mir and keywords.mir: change type ids
-        return_code, unknown_types = CmdCopy._change_type_ids(single_task_annotations=single_task_annotations,
-                                                              data_mir_root=data_mir_root,
-                                                              dst_mir_root=dst_mir_root)
-        if return_code != MirCode.RC_OK:
-            raise MirRuntimeError(error_code=return_code, error_message='change annotation type ids failed')
-        if unknown_types:
-            if ignore_unknown_types:
-                logging.warning(f"unknown types: {unknown_types}")
-            else:
-                raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_MIR_REPO,
-                                      error_message=f"copy annotations error, unknown types: {unknown_types}")
-
-        return single_task_annotations, unknown_types
+        unknown_names_and_count: Dict[str, int] = {}
+        for src_id in unknown_src_class_ids:
+            name = src_class_id_mgr.main_name_for_id(src_id)
+            cnt_gt: int = mir_context.pred_stats.class_ids_cnt[src_id]
+            cnt_pred: int = mir_context.gt_stats.class_ids_cnt[src_id]
+            unknown_names_and_count[name] = cnt_gt + cnt_pred
+        return unknown_names_and_count
 
 
 def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: argparse.ArgumentParser) -> None:
@@ -206,6 +179,11 @@ def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: ar
                                  type=str,
                                  required=True,
                                  help="type:rev@bid, branch in source mir root")
+    copy_arg_parser.add_argument("--src-user-label-file",
+                                 dest="data_label_storage_file",
+                                 type=str,
+                                 required=False,
+                                 help="source label storage file path you want to copy from")
     copy_arg_parser.add_argument("--dst-rev", dest="dst_rev", type=str, required=True, help="rev@tid")
     copy_arg_parser.add_argument('-w', dest='work_dir', type=str, required=False, help='working directory')
     copy_arg_parser.add_argument('--ignore-unknown-types',
