@@ -8,6 +8,7 @@ from mir.commands import base
 from mir.protos import mir_command_pb2 as mirpb
 from mir.scm.cmd import CmdScm
 from mir.tools import checker, class_ids, mir_storage_ops, revs_parser
+from mir.tools.annotations import make_empty_mir_annotations
 from mir.tools.code import MirCode
 
 
@@ -28,38 +29,41 @@ class CmdInit(base.BaseCommand):
 
         dst_rev_tid = revs_parser.parse_single_arg_rev(empty_rev, need_tid=True)
 
-        mir_metadatas = mirpb.MirMetadatas()
-        mir_annotations = mirpb.MirAnnotations()
         task = mir_storage_ops.create_task(task_type=mirpb.TaskTypeInit,
                                            task_id=dst_rev_tid.tid,
                                            message='init empty dataset',
                                            src_revs='master',
                                            dst_rev=empty_rev)
-        mir_datas = {
-            mirpb.MirStorage.MIR_METADATAS: mir_metadatas,
-            mirpb.MirStorage.MIR_ANNOTATIONS: mir_annotations,
-        }
         mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                       mir_branch=dst_rev_tid.rev,
                                                       his_branch='master',
-                                                      mir_datas=mir_datas,
+                                                      mir_datas={
+                                                          mirpb.MirStorage.MIR_METADATAS: mirpb.MirMetadatas(),
+                                                          mirpb.MirStorage.MIR_ANNOTATIONS:
+                                                          make_empty_mir_annotations()
+                                                      },
                                                       task=task)
 
     # public: run
     @staticmethod
-    def run_with_args(mir_root: str, empty_rev: str) -> int:
+    def run_with_args(mir_root: str, label_storage_file: str, empty_rev: str) -> int:
         return_code = checker.check(
             mir_root, [checker.Prerequisites.IS_OUTSIDE_GIT_REPO, checker.Prerequisites.IS_OUTSIDE_MIR_REPO])
         if return_code != MirCode.RC_OK:
             return return_code
 
-        class_ids.create_empty_if_not_exists(mir_root=mir_root)
+        class_ids.load_or_create_userlabels(label_storage_file=label_storage_file, create_ok=True)
 
         repo_git = scm.Scm(root_dir=mir_root, scm_executable='git')
         repo_git.init()
         repo_git.config(['core.fileMode', 'false'])
 
         CmdInit.__update_ignore(mir_root=mir_root, git=repo_git, ignored_items=['.mir_lock', '.mir'])
+
+        with open(os.path.join(mir_root, '.gitattributes'), 'w') as f:
+            f.write('*.mir binary\n')
+        repo_git.add('.')
+
         repo_git.commit(["-m", "first commit"])
 
         # creates an empty dataset if empty_rev provided
@@ -71,6 +75,7 @@ class CmdInit(base.BaseCommand):
         logging.debug("command init: %s", self.args)
 
         return self.run_with_args(mir_root=self.args.mir_root,
+                                  label_storage_file=self.args.label_storage_file,
                                   empty_rev=self.args.empty_rev)
 
 
