@@ -3,14 +3,14 @@ import json
 import logging
 import os
 from subprocess import CalledProcessError
-from typing import Dict, Optional, Set
-
-from google.protobuf import json_format
-import yaml
+from typing import Optional, Set
 
 from mir.commands import base, infer
+from mir.commands.merge import merge_with_pb
 from mir.protos import mir_command_pb2 as mirpb
-from mir.tools import checker, class_ids, data_exporter, mir_storage_ops, revs_parser, utils as mir_utils
+from mir.tools import checker, class_ids, env_config, exporter
+from mir.tools import mir_storage_ops, models, revs_parser
+from mir.tools.annotations import filter_mirdatas_by_asset_ids, MergeStrategy
 from mir.tools.code import MirCode
 from mir.tools.command_run_in_out import command_run_in_out
 from mir.tools.errors import MirContainerError, MirRuntimeError
@@ -33,32 +33,38 @@ class CmdMining(base.BaseCommand):
         return CmdMining.run_with_args(work_dir=self.args.work_dir,
                                        asset_cache_dir=self.args.asset_cache_dir,
                                        src_revs=self.args.src_revs,
+                                       strategy=MergeStrategy(self.args.strategy),
                                        dst_rev=self.args.dst_rev,
                                        mir_root=self.args.mir_root,
-                                       model_hash=self.args.model_hash,
+                                       label_storage_file=self.args.label_storage_file,
+                                       model_hash_stage=self.args.model_hash_stage,
                                        model_location=self.args.model_location,
                                        media_location=self.args.media_location,
                                        config_file=self.args.config_file,
                                        topk=self.args.topk,
-                                       add_annotations=self.args.add_annotations,
+                                       add_prediction=self.args.add_prediction,
                                        executor=self.args.executor,
-                                       executant_name=self.args.executant_name)
+                                       executant_name=self.args.executant_name,
+                                       run_as_root=self.args.run_as_root)
 
     @staticmethod
     @command_run_in_out
     def run_with_args(work_dir: str,
                       asset_cache_dir: Optional[str],
                       src_revs: str,
+                      strategy: MergeStrategy,
                       dst_rev: str,
                       mir_root: str,
-                      model_hash: str,
+                      label_storage_file: str,
+                      model_hash_stage: str,
                       media_location: str,
                       model_location: str,
                       config_file: str,
                       executor: str,
                       executant_name: str,
+                      run_as_root: bool,
                       topk: int = None,
-                      add_annotations: bool = False) -> int:
+                      add_prediction: bool = False) -> int:
         """
         runs a mining task \n
         Args:
@@ -67,13 +73,13 @@ class CmdMining(base.BaseCommand):
             src_revs: data branch name and base task id
             dst_rev: destination branch name and task id
             mir_root: mir repo path, in order to run in non-mir folder.
-            model_hash: used to target model, use prep_tid if non-set
+            model_hash_stage: model_hash@stage_name
             media_location, model_location: location of assets.
             config_file: path to the config file
             executor: executor name, currently, the docker image name
             executant_name: docker container name
             topk: top k assets you want to select in the result workspace, positive integer or None (no mining)
-            add_annotations: if true, write new annotations into annotations.mir
+            add_prediction: if true, write new prediction into annotations.mir
         Returns:
             error code
         """
@@ -85,11 +91,11 @@ class CmdMining(base.BaseCommand):
         if not media_location or not model_location:
             logging.error('media or model location cannot be none!')
             return MirCode.RC_CMD_INVALID_ARGS
-        if not model_hash:
-            logging.error('model_hash is required.')
+        if not model_hash_stage:
+            logging.error('model_hash_stage is required.')
             return MirCode.RC_CMD_INVALID_ARGS
 
-        src_typ_rev_tid = revs_parser.parse_single_arg_rev(src_revs, need_tid=False)
+        src_typ_rev_tids = revs_parser.parse_arg_revs(src_revs)
         dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev, need_tid=True)
 
         if not config_file:
@@ -108,11 +114,10 @@ class CmdMining(base.BaseCommand):
             return return_code
 
         # check `topk` and `add_annotations`
-        mir_metadatas: mirpb.MirMetadatas = mir_storage_ops.MirStorageOps.load_single_storage(
-            mir_root=mir_root,
-            mir_branch=src_typ_rev_tid.rev,
-            mir_task_id=src_typ_rev_tid.tid,
-            ms=mirpb.MirStorage.MIR_METADATAS)
+        mir_metadatas, mir_annotations = merge_with_pb(mir_root=mir_root,
+                                                       src_typ_rev_tids=src_typ_rev_tids,
+                                                       ex_typ_rev_tids=[],
+                                                       strategy=strategy)
         assets_count = len(mir_metadatas.attributes)
         if assets_count == 0:
             raise MirRuntimeError(error_code=MirCode.RC_CMD_MERGE_ERROR,
@@ -141,39 +146,51 @@ class CmdMining(base.BaseCommand):
             logging.error('mining enviroment prepare error!')
             return ret
 
-        _prepare_assets(mir_metadatas=mir_metadatas,
-                        mir_root=mir_root,
-                        src_rev_tid=src_typ_rev_tid,
-                        media_location=media_location,
-                        work_asset_path=work_asset_path,
-                        work_index_file=work_index_file)
+        # export assets.
+        # mining export abs assets path, which will be converted in-docker path in infer.py.
+        ec = mirpb.ExportConfig(asset_format=mirpb.AssetFormat.AF_RAW,
+                                asset_dir=work_asset_path,
+                                asset_index_file=work_index_file,
+                                media_location=media_location,
+                                need_sub_folder=True,
+                                anno_format=mirpb.ExportFormat.EF_NO_ANNOTATIONS,)
+        export_code = exporter.export_mirdatas_to_dir(
+            mir_metadatas=mir_metadatas,
+            ec=ec,
+        )
+        if export_code != MirCode.RC_OK:
+            return export_code
+
+        model_hash, stage_name = models.parse_model_hash_stage(model_hash_stage)
+        model_storage = models.prepare_model(model_location=model_location,
+                                             model_hash=model_hash,
+                                             stage_name=stage_name,
+                                             dst_model_path=work_model_path)
 
         return_code = MirCode.RC_OK
         return_msg = ''
         try:
-            infer.CmdInfer.run_with_args(work_dir=work_dir,
-                                         mir_root=mir_root,
-                                         media_path=work_asset_path,
-                                         model_location=model_location,
-                                         model_hash=model_hash,
-                                         index_file=work_index_file,
-                                         config_file=config_file,
-                                         task_id=dst_typ_rev_tid.tid,
-                                         shm_size=_get_shm_size(config_file),
-                                         executor=executor,
-                                         executant_name=executant_name,
-                                         run_infer=add_annotations,
-                                         run_mining=(topk is not None))
+            return_code = infer.CmdInfer.run_with_args(work_dir=work_dir,
+                                                       label_storage_file=label_storage_file,
+                                                       media_path=work_asset_path,
+                                                       model_storage=model_storage,
+                                                       index_file=work_index_file,
+                                                       config_file=config_file,
+                                                       task_id=dst_typ_rev_tid.tid,
+                                                       executor=executor,
+                                                       executant_name=executant_name,
+                                                       run_as_root=run_as_root,
+                                                       run_infer=add_prediction,
+                                                       run_mining=(topk is not None))
         except CalledProcessError:
             return_code = MirCode.RC_CMD_CONTAINER_ERROR
-            return_msg = mir_utils.collect_executor_outlog_tail(work_dir=work_dir)
+            return_msg = env_config.collect_executor_outlog_tail(work_dir=work_dir)
         # catch other exceptions in command_run_in_out
 
         task = mir_storage_ops.create_task(task_type=mirpb.TaskTypeMining,
                                            task_id=dst_typ_rev_tid.tid,
-                                           message='mining',
-                                           model_hash=model_hash,
-                                           src_revs=src_typ_rev_tid.rev_tid,
+                                           message=f"mining with model: {model_hash_stage}",
+                                           src_revs=src_revs,
                                            dst_rev=dst_typ_rev_tid.rev_tid,
                                            return_code=return_code,
                                            return_msg=return_msg,
@@ -182,77 +199,73 @@ class CmdMining(base.BaseCommand):
             raise MirContainerError(error_message='mining container error occured', task=task)
 
         _process_results(mir_root=mir_root,
+                         label_storage_file=label_storage_file,
                          export_out=work_out_path,
-                         dst_typ_rev_tid=dst_typ_rev_tid,
-                         src_typ_rev_tid=src_typ_rev_tid,
                          topk=topk,
-                         add_annotations=add_annotations,
-                         task=task)
+                         add_prediction=add_prediction,
+                         model_storage=model_storage,
+                         mir_metadatas=mir_metadatas,
+                         mir_annotations=mir_annotations)
+
+        mir_datas = {
+            mirpb.MirStorage.MIR_METADATAS: mir_metadatas,
+            mirpb.MirStorage.MIR_ANNOTATIONS: mir_annotations,
+        }
+        mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
+                                                      his_branch=src_typ_rev_tids[0].rev,
+                                                      mir_branch=dst_typ_rev_tid.rev,
+                                                      mir_datas=mir_datas,
+                                                      task=task)
         logging.info(f"mining done, results at: {work_out_path}")
 
         return MirCode.RC_OK
 
 
 # protected: post process
-def _process_results(mir_root: str, export_out: str, dst_typ_rev_tid: revs_parser.TypRevTid,
-                     src_typ_rev_tid: revs_parser.TypRevTid, topk: Optional[int], add_annotations: bool,
-                     task: mirpb.Task) -> int:
+def _process_results(mir_root: str, label_storage_file: str, export_out: str,
+                     topk: Optional[int], add_prediction: bool,
+                     model_storage: models.ModelStorage, mir_metadatas: mirpb.MirMetadatas,
+                     mir_annotations: mirpb.MirAnnotations) -> None:
     # step 1: build topk results:
-    #   read old
-    [mir_metadatas, mir_annotations] = mir_storage_ops.MirStorageOps.load_multiple_storages(
-        mir_root=mir_root,
-        mir_branch=src_typ_rev_tid.rev,
-        mir_task_id=src_typ_rev_tid.tid,
-        ms_list=[mirpb.MirStorage.MIR_METADATAS, mirpb.MirStorage.MIR_ANNOTATIONS],
-        as_dict=False,
-    )
-
     #   parse new: topk and new annotations (both optional)
     topk_result_file_path = os.path.join(export_out, 'result.tsv')
     asset_ids_set = (_get_topk_asset_ids(file_path=topk_result_file_path, topk=topk)
                      if topk is not None else set(mir_metadatas.attributes.keys()))
 
-    infer_result_file_path = os.path.join(export_out, 'infer-result.json')
-    cls_id_mgr = class_ids.ClassIdManager(mir_root=mir_root)
-    asset_id_to_annotations = (_get_infer_annotations(file_path=infer_result_file_path,
-                                                      asset_ids_set=asset_ids_set,
-                                                      cls_id_mgr=cls_id_mgr) if add_annotations else {})
+    # step 2: filter by topk asset ids
+    filter_mirdatas_by_asset_ids(mir_metadatas=mir_metadatas,
+                                 mir_annotations=mir_annotations,
+                                 asset_ids_set=asset_ids_set)
 
-    # step 2: update mir data files
-    #   update mir metadatas
-    matched_mir_metadatas = mirpb.MirMetadatas()
-    for asset_id in asset_ids_set:
-        matched_mir_metadatas.attributes[asset_id].CopyFrom(mir_metadatas.attributes[asset_id])
-    logging.info(f"matched: {len(matched_mir_metadatas.attributes)}, overriding metadatas.mir")
+    # step 3: if necessary, add predictions from infer result file
+    #   prediction.mir is generated by cmd infer
+    if add_prediction:
+        prediction = mir_annotations.prediction
+        prediction.Clear()
 
-    #   update mir annotations
-    matched_mir_annotations = mirpb.MirAnnotations()
-    matched_task_annotation = matched_mir_annotations.task_annotations[dst_typ_rev_tid.tid]
-    if add_annotations:
-        # add new
-        for asset_id, single_image_annotations in asset_id_to_annotations.items():
-            matched_task_annotation.image_annotations[asset_id].CopyFrom(single_image_annotations)
-    else:
-        # use old
-        task_annotation = mir_annotations.task_annotations[mir_annotations.head_task_id]
-        joint_asset_ids_set = set(task_annotation.image_annotations.keys()) & asset_ids_set
-        for asset_id in joint_asset_ids_set:
-            matched_task_annotation.image_annotations[asset_id].CopyFrom(task_annotation.image_annotations[asset_id])
+        cls_id_mgr = class_ids.load_or_create_userlabels(label_storage_file=label_storage_file)
 
-    #   mir_keywords: auto generated from mir_annotations, so do nothing
+        infer_result_prediction = mirpb.SingleTaskAnnotations()
+        with open(os.path.join(export_out, 'prediction.mir'), 'rb') as f:
+            infer_result_prediction.ParseFromString(f.read())
 
-    # step 3: store results and commit.
-    mir_datas = {
-        mirpb.MirStorage.MIR_METADATAS: matched_mir_metadatas,
-        mirpb.MirStorage.MIR_ANNOTATIONS: matched_mir_annotations,
-    }
-    mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
-                                                  his_branch=src_typ_rev_tid.rev,
-                                                  mir_branch=dst_typ_rev_tid.rev,
-                                                  mir_datas=mir_datas,
-                                                  task=task)
+        # filter and convert
+        #   filter: we need only asset ids in asset_ids_set
+        #   convert: cmd infer packs infer_result_prediction.image_annotations with file base name as key
+        #       so we need to convert all keys to asset hash (in cmd mining, the image's main name)
+        for file_name in infer_result_prediction.image_annotations:
+            asset_id = os.path.splitext(file_name)[0]
+            if asset_id in asset_ids_set:
+                prediction.image_annotations[asset_id].CopyFrom(
+                    infer_result_prediction.image_annotations[file_name])
 
-    return MirCode.RC_OK
+        # pred type and meta
+        prediction.eval_class_ids[:] = set(
+            cls_id_mgr.id_for_names(model_storage.class_names, drop_unknown_names=True)[0])
+        prediction.executor_config = json.dumps(model_storage.executor_config)
+        prediction.model.CopyFrom(model_storage.get_model_meta())
+        prediction.type = (mirpb.ObjectType.OT_DET_BOX
+                           if model_storage.object_type == mirpb.ObjectType.OT_DET_BOX else mirpb.ObjectType.OT_SEG)
 
 
 def _get_topk_asset_ids(file_path: str, topk: int) -> Set[str]:
@@ -274,43 +287,6 @@ def _get_topk_asset_ids(file_path: str, topk: int) -> Set[str]:
     return asset_ids_set
 
 
-def _get_infer_annotations(file_path: str, asset_ids_set: Set[str],
-                           cls_id_mgr: class_ids.ClassIdManager) -> Dict[str, mirpb.SingleImageAnnotations]:
-    asset_id_to_annotations: dict = {}
-    with open(file_path, 'r') as f:
-        results = json.loads(f.read())
-
-    if 'detection' not in results or not isinstance(results['detection'], dict):
-        logging.error('invalid infer-result.json')
-        return asset_id_to_annotations
-
-    names_annotations_dict = results['detection']
-    for asset_name, annotations_dict in names_annotations_dict.items():
-        if 'annotations' not in annotations_dict or not isinstance(annotations_dict['annotations'], list):
-            continue
-        asset_id = os.path.splitext(os.path.basename(asset_name))[0]
-        if asset_id not in asset_ids_set:
-            logging.info(f"unknown asset name: {asset_name}, ignore")
-            continue
-        single_image_annotations = mirpb.SingleImageAnnotations()
-        idx = 0
-        for annotation_dict in annotations_dict['annotations']:
-            class_id = cls_id_mgr.id_and_main_name_for_name(name=annotation_dict['class_name'])[0]
-            # ignore unknown class ids
-            if class_id < 0:
-                continue
-
-            annotation = mirpb.Annotation()
-            annotation.index = idx
-            json_format.ParseDict(annotation_dict['box'], annotation.box)
-            annotation.class_id = class_id
-            annotation.score = float(annotation_dict.get('score', 0))
-            single_image_annotations.annotations.append(annotation)
-            idx += 1
-        asset_id_to_annotations[asset_id] = single_image_annotations
-    return asset_id_to_annotations
-
-
 # protected: pre process
 def _prepare_env(export_root: str, work_in_path: str, work_out_path: str, work_asset_path: str,
                  work_model_path: str) -> int:
@@ -323,32 +299,6 @@ def _prepare_env(export_root: str, work_in_path: str, work_out_path: str, work_a
     return MirCode.RC_OK
 
 
-def _prepare_assets(mir_metadatas: mirpb.MirMetadatas, mir_root: str, src_rev_tid: revs_parser.TypRevTid,
-                    media_location: str, work_asset_path: str, work_index_file: str) -> None:
-    img_list = set(mir_metadatas.attributes.keys())
-    data_exporter.export(mir_root=mir_root,
-                         assets_location=media_location,
-                         class_type_ids={},
-                         asset_ids=img_list,
-                         asset_dir=work_asset_path,
-                         annotation_dir='',
-                         need_ext=True,
-                         need_id_sub_folder=True,
-                         base_branch=src_rev_tid.rev,
-                         base_task_id=src_rev_tid.tid,
-                         format_type=data_exporter.ExportFormat.EXPORT_FORMAT_NO_ANNOTATION,
-                         index_file_path=work_index_file,
-                         index_assets_prefix=work_asset_path)
-
-
-def _get_shm_size(mining_config_file_path: str) -> str:
-    with open(mining_config_file_path, 'r') as f:
-        mining_config = yaml.safe_load(f.read())
-    if 'shm_size' not in mining_config:
-        return '16G'
-    return mining_config['shm_size']
-
-
 # public: arg parser
 def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: argparse.ArgumentParser) -> None:
     mining_arg_parser = subparsers.add_parser('mining',
@@ -380,21 +330,28 @@ def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: ar
                                    type=int,
                                    required=False,
                                    help='if set, discard samples out of topk, sorting by scores.')
-    mining_arg_parser.add_argument('--add-annotations',
-                                   dest='add_annotations',
+    mining_arg_parser.add_argument('--add-prediction',
+                                   dest='add_prediction',
                                    action='store_true',
                                    required=False,
-                                   help='if set, also add inference result to annotations')
+                                   help='if set, also add inference result')
     mining_arg_parser.add_argument('--model-hash',
-                                   dest='model_hash',
+                                   dest='model_hash_stage',
                                    type=str,
                                    required=True,
-                                   help='model hash to be used')
+                                   help='model hash@stage to be used')
     mining_arg_parser.add_argument('--src-revs',
                                    dest='src_revs',
                                    type=str,
                                    required=True,
-                                   help='rev@bid: source rev and base task id')
+                                   help="source revs and base task ids, first the host, others the guests")
+    mining_arg_parser.add_argument("-s",
+                                   dest="strategy",
+                                   type=str,
+                                   default="stop",
+                                   choices=["stop", "host", "guest"],
+                                   help="conflict resolvation strategy, stop (default): stop when conflict detects; "
+                                   "host: use host; guest: use guest")
     mining_arg_parser.add_argument('--dst-rev',
                                    dest='dst_rev',
                                    type=str,
@@ -415,4 +372,8 @@ def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: ar
                                    dest='executant_name',
                                    type=str,
                                    help='docker container name for mining')
+    mining_arg_parser.add_argument("--run-as-root",
+                                   dest="run_as_root",
+                                   action='store_true',
+                                   help="run executor as root user")
     mining_arg_parser.set_defaults(func=CmdMining)
