@@ -2,97 +2,85 @@ import argparse
 import logging
 import os
 import time
-import subprocess
 from subprocess import CalledProcessError
-import traceback
-from typing import Any, List, Optional, Set, Tuple
+from typing import Any, Dict, List, Optional
+from mir.version import YMIR_MODEL_VERSION
 
 from tensorboardX import SummaryWriter
 import yaml
 
 from mir.commands import base
+from mir.commands.merge import merge_with_pb
 from mir.protos import mir_command_pb2 as mirpb
-from mir.tools import checker, class_ids, context, data_exporter, mir_storage_ops, revs_parser
-from mir.tools import settings as mir_settings, utils as mir_utils
+from mir.tools import checker, class_ids, env_config, exporter
+from mir.tools import mir_storage_ops, models, revs_parser
+from mir.tools import settings as mir_settings
+from mir.tools.annotations import make_empty_mir_annotations, MergeStrategy
 from mir.tools.command_run_in_out import command_run_in_out
 from mir.tools.code import MirCode
 from mir.tools.errors import MirContainerError, MirRuntimeError
+from mir.tools.executant import prepare_executant_env, run_docker_executant
 
 
 # private: post process
-def _process_model_storage(out_root: str, model_upload_location: str, executor_config: dict,
-                           task_context: dict) -> Tuple[str, float]:
+def _get_model_storage(model_root: str, executor_config: dict, task_context: dict) -> models.ModelStorage:
     """
-    find and save models
-    Returns:
-        model hash, model mAP and ModelStorage
-    """
-    out_model_dir = os.path.join(out_root, "models")
-    model_paths, model_mAP = _find_models(out_model_dir)
-    if not model_paths:
-        # if have no models
-        return '', model_mAP
-
-    model_storage = mir_utils.ModelStorage(executor_config=executor_config,
-                                           task_context=dict(**task_context,
-                                                             mAP=model_mAP,
-                                                             type=mirpb.TaskType.TaskTypeTraining),
-                                           models=[os.path.basename(model_path) for model_path in model_paths])
-    model_sha1 = mir_utils.pack_and_copy_models(model_storage=model_storage,
-                                                model_dir_path=out_model_dir,
-                                                model_location=model_upload_location)
-
-    return model_sha1, model_mAP
-
-
-def _find_models(model_root: str) -> Tuple[List[str], float]:
-    """
-    find models in `model_root`, and returns model names and mAP
+    find models in `model_root`, and returns all model stages and attachments
 
     Args:
         model_root (str): model root
 
     Returns:
-        Tuple[List[str], float]: list of model names and map
+        models.ModelStorage
     """
-    model_names = []
-    model_mAP = 0.0
+    # model_names = []
+    # model_mAP = 0.0
+    model_stages: Dict[str, models.ModelStageStorage]
+    best_stage_name = ''
+    attachments: Dict[str, Any] = {}
 
     result_yaml_path = os.path.join(model_root, "result.yaml")
     try:
         with open(result_yaml_path, "r") as f:
             yaml_obj = yaml.safe_load(f.read())
+        if 'model' in yaml_obj:
+            # old trainig result file: read models from `model` field
             model_names = yaml_obj["model"]
             model_mAP = float(yaml_obj["map"])
+
+            best_stage_name = 'default_best_stage'
+            model_stages = {
+                best_stage_name:
+                models.ModelStageStorage(stage_name=best_stage_name,
+                                         files=model_names,
+                                         mAP=model_mAP,
+                                         timestamp=int(time.time()))
+            }
+        elif 'model_stages' in yaml_obj:
+            # new training result file: read from model stages
+            model_stages = {k: models.ModelStageStorage.parse_obj(v) for k, v in yaml_obj['model_stages'].items()}
+
+            best_stage_name = yaml_obj['best_stage_name']
+        if 'attachments' in yaml_obj:
+            attachments = yaml_obj['attachments']
     except FileNotFoundError:
-        logging.warning(traceback.format_exc())
-        return [], 0.0
+        error_message = f"can not find file: {result_yaml_path}, executor may have errors, see ymir-executor-out.log"
+        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_FILE, error_message=error_message)
 
-    return ([os.path.join(model_root, os.path.basename(name)) for name in model_names], model_mAP)
+    if not model_stages:
+        raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
+                              error_message='can not find model stages in result.yaml')
 
-
-# private: process
-def _run_train_cmd(cmd: List[str], out_log_path: str) -> int:
-    """
-    invoke training command
-
-    Args:
-        cmd (str): command
-        out_log_path (str): path of log file
-
-    Returns:
-        int: MirCode.RC_OK if success
-
-    Raises:
-        Exception: if out_log_path can not open for append, or cmd returned non-zero code
-    """
-    logging.info(f"training with cmd: {cmd}")
-    logging.info(f"out log path: {out_log_path}")
-    with open(out_log_path, 'a') as f:
-        # run and wait, if non-zero value returned, raise
-        subprocess.run(cmd, check=True, stdout=f, stderr=f, text=True)
-
-    return MirCode.RC_OK
+    return models.ModelStorage(executor_config=executor_config,
+                               task_context=dict(**task_context,
+                                                 mAP=model_stages[best_stage_name].mAP,
+                                                 type=mirpb.TaskType.TaskTypeTraining),
+                               stages=model_stages,
+                               best_stage_name=best_stage_name,
+                               object_type=int(yaml_obj.get('object_type', mirpb.ObjectType.OT_DET_BOX)),
+                               attachments=attachments,
+                               evaluate_config=yaml_obj.get('evaluate_config', {}),
+                               package_version=YMIR_MODEL_VERSION)
 
 
 # private: pre process
@@ -112,13 +100,7 @@ def _generate_config(executor_config: Any, out_config_path: str, task_id: str,
     return executor_config
 
 
-def _get_shm_size(executor_config: dict) -> str:
-    if 'shm_size' not in executor_config:
-        return '16G'
-    return executor_config['shm_size']
-
-
-def _prepare_pretrained_models(model_location: str, model_hash: str, dst_model_dir: str) -> List[str]:
+def _prepare_pretrained_models(model_location: str, model_hash_stage: str, dst_model_dir: str) -> List[str]:
     """
     prepare pretrained models
     * extract models to dst_model_dir
@@ -126,19 +108,21 @@ def _prepare_pretrained_models(model_location: str, model_hash: str, dst_model_d
 
     Args:
         model_location (str): model location
-        model_hash (str): model package hash
+        model_hash_stage (str): model package hash
         dst_model_dir (str): dir where you want to extract model files to
 
     Returns:
-        List[str]: model names
+        List[str]: stage_name/model_names
     """
-    if not model_hash:
+    if not model_hash_stage:
         return []
-    model_storage = mir_utils.prepare_model(model_location=model_location,
-                                            model_hash=model_hash,
-                                            dst_model_path=dst_model_dir)
+    model_hash, stage_name = models.parse_model_hash_stage(model_hash_stage)
+    model_storage = models.prepare_model(model_location=model_location,
+                                         model_hash=model_hash,
+                                         stage_name=stage_name,
+                                         dst_model_path=dst_model_dir)
 
-    return model_storage.models
+    return [f"{stage_name}/{file_name}" for file_name in model_storage.stages[stage_name].files]
 
 
 def _get_task_parameters(config: dict) -> str:
@@ -152,34 +136,41 @@ class CmdTrain(base.BaseCommand):
         return CmdTrain.run_with_args(work_dir=self.args.work_dir,
                                       asset_cache_dir=self.args.asset_cache_dir,
                                       model_upload_location=self.args.model_path,
-                                      pretrained_model_hash=self.args.model_hash,
+                                      pretrained_model_hash_stage=self.args.model_hash_stage,
                                       src_revs=self.args.src_revs,
+                                      strategy=MergeStrategy(self.args.strategy),
                                       dst_rev=self.args.dst_rev,
                                       mir_root=self.args.mir_root,
+                                      label_storage_file=self.args.label_storage_file,
                                       media_location=self.args.media_location,
                                       tensorboard_dir=self.args.tensorboard_dir,
                                       executor=self.args.executor,
                                       executant_name=self.args.executant_name,
+                                      run_as_root=self.args.run_as_root,
                                       config_file=self.args.config_file)
 
     @staticmethod
     @command_run_in_out
     def run_with_args(work_dir: str,
-                      asset_cache_dir: Optional[str],
+                      asset_cache_dir: str,
                       model_upload_location: str,
-                      pretrained_model_hash: str,
+                      pretrained_model_hash_stage: str,
                       executor: str,
                       executant_name: str,
                       src_revs: str,
+                      strategy: MergeStrategy,
                       dst_rev: str,
                       config_file: Optional[str],
                       tensorboard_dir: str,
+                      run_as_root: bool,
+                      label_storage_file: str,
                       mir_root: str = '.',
                       media_location: str = '') -> int:
         if not model_upload_location:
             logging.error("empty --model-location, abort")
             return MirCode.RC_CMD_INVALID_ARGS
-        src_typ_rev_tid = revs_parser.parse_single_arg_rev(src_revs, need_tid=False)
+
+        src_typ_rev_tids = revs_parser.parse_arg_revs(src_revs)
         dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev, need_tid=True)
         if not work_dir:
             raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='empty work_dir')
@@ -191,7 +182,7 @@ class CmdTrain(base.BaseCommand):
                                   error_message=f"invalid --cache {config_file}, not an absolute path for directory")
 
         return_code = checker.check(mir_root,
-                                    [checker.Prerequisites.IS_INSIDE_MIR_REPO, checker.Prerequisites.HAVE_LABELS])
+                                    [checker.Prerequisites.IS_INSIDE_MIR_REPO])
         if return_code != MirCode.RC_OK:
             return return_code
 
@@ -221,61 +212,42 @@ class CmdTrain(base.BaseCommand):
         task_id = dst_typ_rev_tid.tid
         if not executant_name:
             executant_name = f"default-training-{task_id}"
-        if not tensorboard_dir:
-            tensorboard_dir = os.path.join(work_dir, 'out', 'tensorboard')
-        asset_dir = asset_cache_dir or os.path.join(work_dir, 'in', 'assets')
-
-        # if have model_hash, export model
-        pretrained_model_names = _prepare_pretrained_models(model_location=model_upload_location,
-                                                            model_hash=pretrained_model_hash,
-                                                            dst_model_dir=os.path.join(work_dir, 'in', 'models'))
-
-        # get train_ids and val_ids
-        train_ids = set()  # type: Set[str]
-        val_ids = set()  # type: Set[str]
-        unused_ids = set()  # type: Set[str]
-        mir_metadatas: mirpb.MirMetadatas = mir_storage_ops.MirStorageOps.load_single_storage(
-            mir_root=mir_root,
-            mir_branch=src_typ_rev_tid.rev,
-            mir_task_id=src_typ_rev_tid.tid,
-            ms=mirpb.MirStorage.MIR_METADATAS)
-        for asset_id, asset_attr in mir_metadatas.attributes.items():
-            if asset_attr.tvt_type == mirpb.TvtTypeTraining:
-                train_ids.add(asset_id)
-            elif asset_attr.tvt_type == mirpb.TvtTypeValidation:
-                val_ids.add(asset_id)
-            else:
-                unused_ids.add(asset_id)
-        if not train_ids:
-            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='no training set')
-        if not val_ids:
-            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='no validation set')
-
-        if not unused_ids:
-            logging.info(f"training: {len(train_ids)}, validation: {len(val_ids)}")
-        else:
-            logging.warning(f"training: {len(train_ids)}, validation: {len(val_ids)}" f"unused: {len(unused_ids)}")
-
-        # export
-        logging.info("exporting assets")
 
+        # setting up paths.
         os.makedirs(work_dir, exist_ok=True)
 
         work_dir_in = os.path.join(work_dir, "in")
-        work_dir_annotations = os.path.join(work_dir_in, 'annotations')
-        os.makedirs(work_dir_annotations, exist_ok=True)
-
         work_dir_out = os.path.join(work_dir, "out")
-        os.makedirs(work_dir_out, exist_ok=True)
-        out_model_dir = os.path.join(work_dir, 'out', 'models')
-        os.makedirs(out_model_dir, exist_ok=True)
+        prepare_executant_env(work_dir_in=work_dir_in,
+                              work_dir_out=work_dir_out,
+                              asset_cache_dir=asset_cache_dir,
+                              tensorboard_dir=tensorboard_dir)
 
-        os.makedirs(asset_dir, exist_ok=True)
-        os.makedirs(tensorboard_dir, exist_ok=True)
+        asset_dir = os.path.join(work_dir_in, 'assets')
+        work_dir_pred = os.path.join(work_dir_in, 'predictions')
+        work_dir_gt = os.path.join(work_dir_in, 'annotations')
+        tensorboard_dir = os.path.join(work_dir_out, 'tensorboard')
 
+        docker_log_dst = os.path.join(tensorboard_dir, "executor.log")
+        docker_log_src = os.path.join(work_dir_out, mir_settings.EXECUTOR_OUTLOG_NAME)
+        open(docker_log_src, 'w').close()
+        os.symlink(docker_log_src, docker_log_dst)
+
+        # if have model_hash_stage, export model
+        pretrained_model_stage_and_names = _prepare_pretrained_models(model_location=model_upload_location,
+                                                                      model_hash_stage=pretrained_model_hash_stage,
+                                                                      dst_model_dir=os.path.join(work_dir_in, 'models'))
+
+        mir_metadatas, mir_annotations = merge_with_pb(mir_root=mir_root,
+                                                       src_typ_rev_tids=src_typ_rev_tids,
+                                                       ex_typ_rev_tids=[],
+                                                       strategy=strategy)
+
+        # export
+        logging.info("exporting assets")
         # type names to type ids
         # ['cat', 'person'] -> [4, 2]
-        cls_mgr = class_ids.ClassIdManager(mir_root=mir_root)
+        cls_mgr = class_ids.load_or_create_userlabels(label_storage_file=label_storage_file)
         type_ids_list, unknown_names = cls_mgr.id_for_names(class_names)
         if not type_ids_list:
             logging.info(f"type ids empty, please check config file: {config_file}")
@@ -284,46 +256,32 @@ class CmdTrain(base.BaseCommand):
             raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS,
                                   error_message=f"unknown class names: {unknown_names}")
 
-        if not context.check_class_ids(mir_root=mir_root, current_class_ids=type_ids_list):
-            raise MirRuntimeError(error_code=MirCode.RC_CMD_INVALID_ARGS, error_message='user class ids mismatch')
-
         type_id_idx_mapping = {type_id: index for (index, type_id) in enumerate(type_ids_list)}
-
-        # export train set
-        data_exporter.export(mir_root=mir_root,
-                             assets_location=media_location,
-                             class_type_ids=type_id_idx_mapping,
-                             asset_ids=train_ids,
-                             asset_dir=asset_dir,
-                             annotation_dir=work_dir_annotations,
-                             need_ext=True,
-                             need_id_sub_folder=True,
-                             base_branch=src_typ_rev_tid.rev,
-                             base_task_id=src_typ_rev_tid.tid,
-                             format_type=data_exporter.ExportFormat.EXPORT_FORMAT_ARK,
-                             index_file_path=os.path.join(work_dir_in, 'train-index.tsv'),
-                             index_assets_prefix='/in/assets',
-                             index_annotations_prefix='/in/annotations')
-
-        # export validation set
-        data_exporter.export(mir_root=mir_root,
-                             assets_location=media_location,
-                             class_type_ids=type_id_idx_mapping,
-                             asset_ids=val_ids,
-                             asset_dir=asset_dir,
-                             annotation_dir=work_dir_annotations,
-                             need_ext=True,
-                             need_id_sub_folder=True,
-                             base_branch=src_typ_rev_tid.rev,
-                             base_task_id=src_typ_rev_tid.tid,
-                             format_type=data_exporter.ExportFormat.EXPORT_FORMAT_ARK,
-                             index_file_path=os.path.join(work_dir_in, 'val-index.tsv'),
-                             index_assets_prefix='/in/assets',
-                             index_annotations_prefix='/in/annotations')
-
-        logging.info("starting train docker container")
-
-        available_gpu_id = config.get(mir_settings.TASK_CONTEXT_KEY, {}).get('available_gpu_id', '')
+        anno_format, asset_format = exporter.parse_export_type(type_str=executor_config.get('export_format', ''))
+        ec = mirpb.ExportConfig(asset_format=asset_format,
+                                asset_dir=asset_dir,
+                                asset_index_file=os.path.join(work_dir_in, "idx-assets.tsv"),
+                                asset_index_prefix="/in/assets",
+                                media_location=media_location,
+                                need_sub_folder=True,
+                                anno_format=anno_format,
+                                gt_dir=work_dir_gt,
+                                gt_index_file=os.path.join(work_dir_in, "idx-gt.tsv"),
+                                gt_index_prefix="/in/annotations",
+                                pred_dir=work_dir_pred,
+                                pred_index_file=os.path.join(work_dir_in, "idx-pred.tsv"),
+                                pred_index_prefix="/in/predictions",
+                                tvt_index_dir=work_dir_in,)
+        export_code = exporter.export_mirdatas_to_dir(
+            mir_metadatas=mir_metadatas,
+            ec=ec,
+            mir_annotations=mir_annotations,
+            class_ids_mapping=type_id_idx_mapping,
+            cls_id_mgr=cls_mgr,
+        )
+        if export_code != MirCode.RC_OK:
+            return export_code
+        logging.info("finish exporting, starting train docker container")
 
         # generate configs
         out_config_path = os.path.join(work_dir_in, "config.yaml")
@@ -331,34 +289,29 @@ class CmdTrain(base.BaseCommand):
             executor_config=executor_config,
             out_config_path=out_config_path,
             task_id=task_id,
-            pretrained_model_params=[os.path.join('/in/models', name) for name in pretrained_model_names])
-        mir_utils.generate_training_env_config_file(task_id=task_id,
-                                                    env_config_file_path=os.path.join(work_dir_in, 'env.yaml'))
-
-        # start train docker and wait
-        path_binds = []
-        path_binds.append(f"-v{work_dir_in}:/in")  # annotations, models, train-index.tsv, val-index.tsv, config.yaml
-        path_binds.append(f"-v{asset_dir}:/in/assets:ro")  # assets
-        path_binds.append(f"-v{work_dir_out}:/out")
-        path_binds.append(f"-v{tensorboard_dir}:/out/tensorboard")
-
-        cmd = ['nvidia-docker', 'run', '--rm', f"--shm-size={_get_shm_size(executor_config=executor_config)}"]
-        cmd.extend(path_binds)
-        if available_gpu_id:
-            cmd.extend(['--gpus', f"\"device={available_gpu_id}\""])
-        cmd.extend(['--user', f"{os.getuid()}:{os.getgid()}"])  # run as current user
-        cmd.extend(['--name', f"{executant_name}"])  # executor name used to stop executor
-        cmd.append(executor)
+            pretrained_model_params=[os.path.join('/in/models', name) for name in pretrained_model_stage_and_names])
+        env_config.generate_training_env_config_file(task_id=task_id,
+                                                     env_config_file_path=os.path.join(work_dir_in, 'env.yaml'))
 
+        task_config = config.get(mir_settings.TASK_CONTEXT_KEY, {})
         task_code = MirCode.RC_OK
-        return_msg = ''
+        return_msg = ""
         try:
-            _run_train_cmd(cmd, out_log_path=os.path.join(work_dir_out, mir_settings.EXECUTOR_OUTLOG_NAME))
+            run_docker_executant(
+                work_dir_in=work_dir_in,
+                work_dir_out=work_dir_out,
+                executor=executor,
+                executant_name=executant_name,
+                executor_config=executor_config,
+                gpu_id=task_config.get('available_gpu_id', ""),
+                run_as_root=run_as_root,
+                task_config=task_config,
+            )
         except CalledProcessError as e:
             logging.warning(f"training exception: {e}")
             # don't exit, proceed if model exists
             task_code = MirCode.RC_CMD_CONTAINER_ERROR
-            return_msg = mir_utils.collect_executor_outlog_tail(work_dir=work_dir)
+            return_msg = env_config.collect_executor_outlog_tail(work_dir=work_dir)
 
             # write executor tail to tensorboard
             if return_msg:
@@ -366,27 +319,30 @@ class CmdTrain(base.BaseCommand):
                     tb_writer.add_text(tag='executor tail', text_string=f"```\n{return_msg}\n```", walltime=time.time())
 
         # gen task_context
-        task_context = {
+        task_context = task_config
+        task_context.update({
             'src_revs': src_revs,
             'dst_rev': dst_rev,
             'executor': executor,
             mir_settings.PRODUCER_KEY: mir_settings.PRODUCER_NAME,
             mir_settings.TASK_CONTEXT_PARAMETERS_KEY: task_parameters
-        }
+        })
 
         # save model
-        logging.info("saving models")
-        model_sha1, model_mAP = _process_model_storage(out_root=work_dir_out,
-                                                       model_upload_location=model_upload_location,
-                                                       executor_config=executor_config,
-                                                       task_context=task_context)
+        logging.info(f"saving models:\n task_context: {task_context}")
+        out_model_dir = os.path.join(work_dir_out, "models")
+        model_storage = _get_model_storage(model_root=out_model_dir,
+                                           executor_config=executor_config,
+                                           task_context=task_context)
+        models.pack_and_copy_models(model_storage=model_storage,
+                                    model_dir_path=out_model_dir,
+                                    model_location=model_upload_location)
 
         # commit task
         task = mir_storage_ops.create_task(task_type=mirpb.TaskType.TaskTypeTraining,
                                            task_id=dst_typ_rev_tid.tid,
                                            message='training',
-                                           model_mAP=model_mAP,
-                                           model_hash=model_sha1,
+                                           model_meta=model_storage.get_model_meta(),
                                            return_code=task_code,
                                            return_msg=return_msg,
                                            serialized_task_parameters=task_parameters,
@@ -400,8 +356,12 @@ class CmdTrain(base.BaseCommand):
 
         mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                       mir_branch=dst_typ_rev_tid.rev,
-                                                      his_branch=src_typ_rev_tid.rev,
-                                                      mir_datas={},
+                                                      his_branch=src_typ_rev_tids[0].rev,
+                                                      mir_datas={
+                                                          mirpb.MirStorage.MIR_METADATAS: mirpb.MirMetadatas(),
+                                                          mirpb.MirStorage.MIR_ANNOTATIONS:
+                                                          make_empty_mir_annotations()
+                                                      },
                                                       task=task)
 
         logging.info("training done")
@@ -425,7 +385,7 @@ def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: ar
                                   type=str,
                                   help="media storage location for models")
     train_arg_parser.add_argument('--model-hash',
-                                  dest='model_hash',
+                                  dest='model_hash_stage',
                                   type=str,
                                   required=False,
                                   help='model hash to be used')
@@ -434,6 +394,7 @@ def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: ar
                                   required=False,
                                   dest='asset_cache_dir',
                                   type=str,
+                                  default='',
                                   help='asset cache directory')
     train_arg_parser.add_argument("--executor",
                                   required=True,
@@ -449,7 +410,15 @@ def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: ar
                                   dest="src_revs",
                                   type=str,
                                   required=True,
-                                  help="rev@bid: source rev and base task id")
+                                  help="source tvt types, revs and base task ids, first the host, others the guests, "
+                                  "can begin with tr:/va:/te:, uses own tvt type if no prefix assigned")
+    train_arg_parser.add_argument("-s",
+                                  dest="strategy",
+                                  type=str,
+                                  default="stop",
+                                  choices=["stop", "host", "guest"],
+                                  help="conflict resolvation strategy, stop (default): stop when conflict detects; "
+                                  "host: use host; guest: use guest")
     train_arg_parser.add_argument("--dst-rev",
                                   dest="dst_rev",
                                   type=str,
@@ -465,4 +434,8 @@ def bind_to_subparsers(subparsers: argparse._SubParsersAction, parent_parser: ar
                                   type=str,
                                   required=False,
                                   help="tensorboard log directory")
+    train_arg_parser.add_argument("--run-as-root",
+                                  dest="run_as_root",
+                                  action='store_true',
+                                  help="run executor as root user")
     train_arg_parser.set_defaults(func=CmdTrain)
